from . import TrigFunc
from typing import Any


class Triggon:
    def __init__(
        self, label: str | dict[str, Any], /, new: Any=None, 
        *, debug: bool=False,
    ) -> None: ...

    def set_trigger(
        self, label: str | list[str] | tuple[str, ...], /,
    ) -> None: ...

    def alter_literal(
        self, label: str, /, org: Any, *, index: int=None,
    ) -> Any: ...

    def alter_var(
          self, label: str | dict[str, Any], var: Any=None, /, 
          *, index: int=None,
    ) -> None | Any: ...

    def revert(
            self, label: str | list[str] | tuple[str, ...], /, 
            *, disable: bool=False,
    ) -> None: ...

    def exit_point(self, label: str, func: TrigFunc, /) -> None | Any: ...    

    def trigger_return(
        self, label: str, /, *, index: int=None, do_print: bool=False,
    ) -> None | Any: ...

    def trigger_func(self, label: str, func: TrigFunc, /) -> None | Any: ...


import ast
from typing import Any


SYMBOL = "*"

LABEL_TYPE_ERROR = (
   "'label' must be a string, or a list/tuple containing only strings."
)

class InvalidArgumentError(ValueError): 
    pass


class NoArgumentsError(Exception): 
    pass


class UnDefinedArgument(Exception): 
    pass


class _ExitEarly(Exception): 
    pass


def _check_label_type(label: Any) -> None:
   if not isinstance(label, str):
      raise InvalidArgumentError("Label must be a string.")


def _compare_value_counts(
      total_values: tuple[Any], needed_indices_count: int,
) -> None:
   # Note: `len(total_values)` is always one more than the max index.
   if len(total_values) <= needed_indices_count:
      raise ValueError(
         f"Please set the value to change at position {needed_indices_count}."
      )
    

def _count_symbol(label: str) -> int:
   if not isinstance(label, str):
      raise InvalidArgumentError("`label` must be a string.")
   
   count = 0

   # count only prefix symbols
   for symbol in label:
      if symbol == SYMBOL:
         count += 1
         continue
      return count


def _handle_arg_types(
      label: Any, value: Any | None, index: int=None, get_type: bool=False,
      ) -> dict[str, Any] | tuple[dict[str, Any], ast.AST]:
   if index is not None and not isinstance(index, int):
      raise InvalidArgumentError(
         "The `index` keyword must be a literal value."
      )
      
   if isinstance(label, str):    
      if label == "":
         raise InvalidArgumentError("Label cannot be empty or blank.")  
      elif isinstance(value, (list, tuple)):               
        if len(value) != 0:
          if get_type:
             return ({label: value}, ast.List) # use ast.List for both list and tuple (same handling). 
          
          return {label: value}    
            
        raise InvalidArgumentError("The contents are empty.")
      
      elif isinstance(value, dict):
         raise InvalidArgumentError(
            "dict is not supported for 'value'."
         )     
      else:
         if get_type:
            return ({label: value}, ast.Name) # use ast.Name for both Name and Attribute (same handling).
         
         return {label: value}
      
   elif isinstance(label, dict):      
      if len(label) == 0:
         raise InvalidArgumentError("The contents are empty.")   
      elif value is None:
         if get_type:
            return (label, ast.Dict)
         
         return label
      
      raise InvalidArgumentError("If a dict is given, set the values in it.")
   
   else:
      raise InvalidArgumentError("Unsupported argument type.")



import inspect
from typing import Any, Callable


class TrigFunc:
  """
  Delays function execution for use with `trigger_return()` or `trigger_func()`.

  Use this wrapper to pass functions without calling them.

  You must assign the instance (e.g., F = TrigFunc()) 
  before calling the trigger function.
  """

  _func: Callable | None

  def __init__(self, func: Callable=None) -> None:
    self._func = func

  def __call__(self, *args, **kwargs) -> "TrigFunc":
    if self._func is None:
      raise ValueError("`func` is None")
    
    return TrigFunc(self._func(*args, **kwargs))
    
  def __getattr__(self, name: str) -> Callable[[], Any]:    
    if self._func is not None:
      target = getattr(self._func, name)
    else:
      frame = inspect.currentframe().f_back
      target = frame.f_locals.get(name) or frame.f_globals.get(name)

    if target is None:
        raise AttributeError(f"'{name}' is not a callable function")
    elif not callable(target):
        return TrigFunc(target)
      
    def _wrapper(*args, **kwargs) -> Callable[[], Any]:
      return lambda: target(*args, **kwargs)
    return _wrapper



from types import FrameType
from typing import Any

from ._exceptions import (
    LABEL_TYPE_ERROR,
    SYMBOL,
    _ExitEarly,
    InvalidArgumentError, 
    _check_label_type,
    _compare_value_counts,
    _count_symbol,
    _handle_arg_types,
)
from . import _debug
from . import _var_analysis
from . import _var_update
from .trig_func import TrigFunc


class Triggon:
    debug: bool
    _debug_var: dict[str, tuple[int, str] | list[tuple[int, str]]]
    _trigger_flag: dict[str, bool]
    _new_value: dict[str, tuple[Any]]
    _org_value: dict[str, tuple[Any]]
    _var_list: dict[str, tuple[str]]
    _disable_label: dict[str, bool]
    _id_list: dict[str, int | tuple[int]]
    _return_value: tuple[bool, Any] | None = None
    _lineno: int = None
    _frame: FrameType = None

    def __init__(
        self, label: str | dict[str, Any], /, new: Any=None, 
        *, debug: bool=False,
    ) -> None:
      """
      Registers labels and their corresponding values.

      Values must be passed in the order of their index positions.
      To treat a sequence as a single value, wrap it in another sequence.

      Accepted formats:
      - label, value
      - label, [value]
      - label, (value,)
      - {label: value}
      - {label: [value]}
      - {label: (value,)}
      """

      self.debug = debug
      self._trigger_flag = {}
      self._new_value = {}
      self._org_value = {}
      self._var_list = {}  
      self._disable_label = {}   
      self._id_list = {}

      change_list = _handle_arg_types(label, new)
      self._scan_dict(change_list)

    def _scan_dict(self, arg_dict: dict[str, Any]) -> None:      
      for key, value in arg_dict.items():          
          label = key.lstrip(SYMBOL)
          index = _count_symbol(key)

          if index != 0:
              raise InvalidArgumentError(
                  f"Please remove the '*' prefix from `{key}`. " 
                  "To specify by index, "
                  "provide the values in index order using a list or tuple."
              )

          try:
            self._new_value[label]

            raise InvalidArgumentError(
              f"`{label}` already exists." 
              "Duplicate labels are not allowed for this function."
            ) 
          except KeyError:
            self._add_new_label(label, value)

    def _add_new_label(self, label: str, value: Any, /) -> None:
      if isinstance(value, (list, tuple)):
        length = len(value)
      else:
        length = 1

      if isinstance(value, list) and length > 1: 
        self._new_value[label] = tuple(value)
      elif isinstance(value, tuple) and length > 1:
        self._new_value[label] = value
      elif isinstance(value, list) and length == 1:
        if isinstance(value[0], (list, tuple)):
          self._new_value[label] = value
        else:
          self._new_value[label] = (value[0],)
      elif isinstance(value, tuple) and length == 1:
        if isinstance(value[0], (list, tuple)):
          self._new_value[label] = value 
        else:
          self._new_value[label] = value
      else:
        # Unwrapped single value
        self._new_value[label] = (value,)
        
      self._trigger_flag[label] = False
      self._disable_label[label] = False

      # Create a list of None values—one for each index of this label
      self._org_value[label] = [None] * length
      self._var_list[label] = [None] * length
      self._id_list[label] = [None] * length

    def set_trigger(
        self, label: str | list[str] | tuple[str, ...], /,
    ) -> None:
      """
      Activates the trigger flag for the given label(s).

      If variables were registered via `alter_var()`, 
      their values will be updated when the flag is activated.
      """
    
      if isinstance(label, (list, tuple)):
        for name in label:
          if not isinstance(name, str):
             raise InvalidArgumentError(LABEL_TYPE_ERROR)

          self._check_label_flag(name)
      elif isinstance(label, str):
        self._check_label_flag(label)       
      else:
        raise InvalidArgumentError(LABEL_TYPE_ERROR)
      
      self._clear_frame()

    def alter_literal(
        self, label: str, /, org: Any, *, index: int=None,
    ) -> Any:
      """
      Changes the value at the specified label and position 
      if the flag is active.

      Only accepts immediate values
      (e.g., literals or expressions).
      """

      _check_label_type(label)
      
      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      if index is None:
        index = _count_symbol(label)
      _compare_value_counts(self._new_value[name], index)

      self._org_value[name][index] = org
      flag = self._trigger_flag[name]

      if self.debug:
        self._get_target_frame("alter_literal")
        self._print_val_debug(name, index, flag, org)
        self._clear_frame()

      if not flag:
        return self._org_value[name][index]

      return self._new_value[name][index]  

    def alter_var(
          self, label: str | dict[str, Any], var: Any=None, /, 
          *, index: int=None,
    ) -> None | Any:
        """
        Change the value of variables associated with the given label 
        if the flag is active.

        Only supports variable references (not literals or expressions).

        Supports updating:
        - Global variables
        - Class attributes (fields)
        """

        (change_list, arg_type) = _handle_arg_types(label, var, index, True)
        init_flag = False

        if len(change_list) == 1:
          # When only one label is provided
          label = next(iter(change_list))
          _check_label_type(label)

          name = label.lstrip(SYMBOL)
          self._check_exist_label(name)

          if index is None:
            index = _count_symbol(label)
          _compare_value_counts(self._new_value[name], index)

          if (
            self._var_list[name][index] is None 
            or self._is_new_var(name, index, var)
          ):
            self._store_org_value(name, index, change_list[label])
 
            self._get_target_frame("alter_var")
            self._lineno = self._frame.f_lineno     

            # Initial process to store argument variables 
            self._init_arg_list(change_list, arg_type, index)
            init_flag = True

          trig_flag = self._trigger_flag[name]
          vars = self._var_list[name][index]

          if not trig_flag:
            if self.debug:
              self._get_target_frame("alter_var")

              self._print_var_debug(
                vars, name, index, trig_flag, change_list[label],
              )   
            self._clear_frame()

            return var
          elif not init_flag:
             return var

          self._update_var_value(vars, self._new_value[name][index])  

          if self.debug:
            self._print_var_debug(
              vars, name, index, trig_flag, change_list[label], 
              self._new_value[name][index], change=True,
            )
          self._clear_frame()

          return var
        else:
           # When multiple labels are provided in a dictionary

          if index is not None:
            raise InvalidArgumentError(
              "Cannot use the `index` keyword with a dictionary. " 
              "Use '*' in the label instead." 
            )
          
          for key, val in change_list.items():
            _check_label_type(key)
            
            name = key.lstrip(SYMBOL)
            index = _count_symbol(key)

            if self._org_value[name][index] is None:
              self._store_org_value(name, index, val)

            if (
               not init_flag
               and (self._var_list[name][index] is None 
               or self._is_new_var(name, index, val))
            ):    
              self._get_target_frame("alter_var")
              self._lineno = self._frame.f_lineno

              # Initial process to store argument variables
              self._init_arg_list(change_list, arg_type)
              init_flag = True
            
            if not init_flag:
              continue

            trig_flag = self._trigger_flag[name]
            vars = self._var_list[name][index]  

            if not trig_flag:
              if self.debug:
                self._get_target_frame("alter_var")
                self._print_var_debug(vars, name, index, trig_flag, val)

              continue          

            self._update_var_value(vars, self._new_value[name][index])

            if self.debug:
              self._get_target_frame("alter_var")

              self._print_var_debug(
                vars, name, index, trig_flag, val, 
                self._new_value[name][index], change=True,
              )
            
          self._clear_frame()

    def revert(
          self, label: str | list[str] | tuple[str, ...], /, 
          *, disable: bool=False,
    ) -> None:
      """
      Revert the trigger flag of the specified label to False.
      If `disable` is set to True, the label will be permanently disabled.
      """

      if isinstance(label, (list, tuple)):
        for name in label:
          if not isinstance(name, str):
            raise InvalidArgumentError(LABEL_TYPE_ERROR)
          
          self._revert_label(name, disable)
      elif isinstance(label, str):
        self._revert_label(label, disable)
      else:
        raise InvalidArgumentError(LABEL_TYPE_ERROR)        

    def _revert_label(self, label: str, disable: bool) -> None:
      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      if not self._trigger_flag[name]:
        return

      if disable:
        state = "disable" # for debug
        self._disable_label[name] = True
      else:
        state = "inactive" # for debug
      self._trigger_flag[name] = False

      self._label_has_var(name, "revert", True)

      if self.debug:
        self._get_target_frame("revert")
        self._print_flag_debug(name, state)    
        self._clear_frame()  
    
    def exit_point(self, label: str, func: TrigFunc, /) -> None | Any:
      """
      Handles an early return triggered by `trigger_return()`,  
      based on the specified label.
      """

      name = label.lstrip(SYMBOL)
      self._check_exist_label(name)

      try:
          return func()
      except _ExitEarly:
          if not self._return_value[0]:
              return self._return_value[1]
          
          print(self._return_value[1])

    def trigger_return(
        self, label: str, /, *, index: int=None, do_print: bool=False,
    ) -> None | Any:
        """
        Executes an early return using the set return value,  
        if the trigger flag is active.

        If `do_print` is True, prints the value with the early return.  
        Raises `InvalidArgumentError` if the value is not a string.
        """

        name = label.lstrip(SYMBOL)
        self._check_exist_label(name)

        if index is None:
           index = _count_symbol(label)

        if not self._trigger_flag[name]:
            return 
            
        if do_print:
            if not isinstance(self._new_value[name][index], str):
              raise InvalidArgumentError(
                 "Expected a value of type `str`, "
                 f"but got `{type(self._new_value[name][index]).__name__}`."
              )         
        self._return_value = (do_print, self._new_value[name][index])

        if self.debug:
          self._get_target_frame("trigger_return")
          self._print_trig_debug(name, "Return")

        raise _ExitEarly 
        
    def trigger_func(self, label: str, func: TrigFunc, /) -> None | Any:
        """
        Executes the given function 
        if the trigger flag for the label is active.
        """

        name = label.lstrip(SYMBOL)
        self._check_exist_label(name)

        if self._trigger_flag[name]:
            if self.debug:
              self._get_target_frame("trigger_func")
              self._print_trig_debug(name, "Trigger a function") 
                    
            return func()


for name, func in vars(_var_analysis).items():
    if callable(func):
        setattr(Triggon, name, func)


for name, func in vars(_debug).items():
    if callable(func):
        setattr(Triggon, name, func)


for name, func in vars(_var_update).items():
    if callable(func):
        setattr(Triggon, name, func)


from typing import Any


LOG_DEBUG = "【Debug】 [Line]: {lineno}, [Function]: `{func}`, "

LOG_FLAG = LOG_DEBUG + "[State]: Label `{label}` is {state}"

LOG_TRIG = LOG_DEBUG + "[Label]: `{label}`, [Action]: {action}"

LOG_VALUE = (
  LOG_DEBUG 
  + "[Label]: `{label}` (index {num}), [Flag]: {state}, [Value]: {value}"
)

LOG_VAR = (
  LOG_DEBUG 
  + "[Label]: `{label}` (index {num}), "
  "[Flag]: {state}, [Variable]: `{var}`, [Value]: {value}"
)


def _get_debug_info(self) -> str:
   # only used debug mode
   func_name = self._frame.f_code.co_name

   if self._lineno is None:
      self._lineno = self._frame.f_lineno

   return func_name

def _get_var_name(
      self, var_ref: tuple[str, ...] | list[tuple[str, ...]], value: Any=None,
) -> str | tuple[list[str], list[Any]]:
   if isinstance(var_ref, list):
      var_names = []
      var_values = []
  
      for i, val in enumerate(var_ref):
         if val[0] != self._lineno:
            continue
         elif len(val) == 2 and self._frame.f_globals[val[1]] == value[i]:
            if len(var_ref) == 3:
              class_name = var_ref[2].__class__.__name__
              var_names.append(f"{class_name}.{val[1]}")
            else:
              var_names.append(val[1])

            var_values.append(value[i])
         elif len(val) == 3 and val[2].__dict__[val[1]] == value[i]:
            if len(var_ref) == 3:
              class_name = var_ref[2].__class__.__name__
              var_names.append(f"{class_name}.{val[1]}")
            else:
              var_names.append(val[1])

            var_values.append(value[i])

      return (var_names, var_values)
   else:
      if len(var_ref) == 3:
         class_name = var_ref[2].__class__.__name__
         return f"{class_name}.{var_ref[1]}"
      
      return var_ref[1]

def _print_flag_debug(self, label: str, state: str, reset: bool=True) -> None:
    func_name = self._get_debug_info()

    print(
        LOG_FLAG.format(
           lineno=self._lineno, func=func_name, label=label, state=state,
        )
    )

    if reset:
       self._clear_frame()

def _print_var_debug(
      self, target_index: tuple[Any] | list[tuple[Any]], 
      label: str, index: int, flag: bool, 
      org_value: Any, new_value: Any=None, change: bool=False,
) -> None:
   func_name = self._get_debug_info()

   if isinstance(org_value, (list, tuple)):
      (var_names, var_values) = self._get_var_name(target_index, org_value)

      for i, var in enumerate(var_names):
         if change:
            value = f"{repr(var_values[i])} -> {repr(new_value)}"
         else:
            value = repr(var_values[i])
            print(
               LOG_VAR.format(
                  lineno=self._lineno, func=func_name, 
                  label=label, num=index, 
                  state=flag, var=var, value=value,
               )
            )
   else:
    var_name = self._get_var_name(target_index)

    if change:
        value = f"{repr(org_value)} -> {repr(new_value)}"
    else:
        value = repr(org_value)

    print(
        LOG_VAR.format(
           lineno=self._lineno, func=func_name, 
           label=label, num=index, 
           state=flag, var=var_name, value=value,
        )
    )  

def _print_trig_debug(self, label: str, action: str) -> None:
    func_name = self._get_debug_info()

    print(
        LOG_TRIG.format(
           lineno=self._lineno, func=func_name, label=label, action=action,
        )
    )
    self._clear_frame()
